# Copyright (c) 2017  Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function
from .document import Index
from .imp import import_object, dir_object
from argparse import ArgumentParser

import atexit
import os
import shutil
import signal
import subprocess
import sys
import yaml

PYDOCMD_CONFIG = 'pydocmd.yml'
parser = ArgumentParser()
parser.add_argument('command', choices=['generate', 'build', 'gh-deploy',
                                        'json', 'new', 'serve', 'simple'])
parser.add_argument('subargs', nargs='...')


def read_config():
  """
  Reads and preprocesses the pydoc-markdown configuration file.
  """

  with open(PYDOCMD_CONFIG) as fp:
    config = yaml.safe_load(fp)
  return default_config(config)


def default_config(config):
  args = parser.parse_args()
  config.setdefault('docs_dir', 'sources')
  config.setdefault('gens_dir', '_build/pydocmd')
  config.setdefault('site_dir', '_build/site')
  if args.command == 'simple':
    config.setdefault('headers', 'markdown')
  else:
    config.setdefault('headers', 'html')
  config.setdefault('theme', 'readthedocs')
  config.setdefault('loader', 'pydocmd.loader.PythonLoader')
  config.setdefault('preprocessor', 'pydocmd.preprocessors.simple.Preprocessor')
  config.setdefault('additional_search_paths', [])
  return config


def write_temp_mkdocs_config(inconf):
  """
  Generates a configuration for MkDocs on-the-fly from the pydoc-markdown
  configuration and makes sure it gets removed when this program exists.
  """
  ignored_keys = ('gens_dir', 'pages', 'headers', 'generate', 'loader',
                  'preprocessor', 'additional_search_paths')

  config = {key: value for key, value in inconf.items() if key not in ignored_keys}
  config['docs_dir'] = inconf['gens_dir']
  if 'pages' in inconf:
      config['nav'] = inconf['pages']

  with open('mkdocs.yml', 'w') as fp:
    yaml.dump(config, fp)

  atexit.register(lambda: os.remove('mkdocs.yml'))


def makedirs(path):
  """
  Create the directory *path* if it does not already exist.
  """

  if not os.path.isdir(path):
    os.makedirs(path)


# Also process all pages to copy files outside of the docs_dir to the gens_dir.
def process_pages(data, gens_dir):
  for key in data:
    filename = data[key]
    if isinstance(filename, str) and '<<' in filename:
      filename, source = filename.split('<<')
      filename, source = filename.rstrip(), source.lstrip()
      outfile = os.path.join(gens_dir, filename)
      makedirs(os.path.dirname(outfile))
      shutil.copyfile(source, outfile)
      data[key] = filename
    elif isinstance(filename, dict):
      process_pages(filename, gens_dir)
    elif isinstance(filename, list):
      [process_pages(x, gens_dir) for x in filename]


def copy_source_files(config, pages_required=True):
  """
  Copies all files from the `docs_dir` to the `gens_dir` defined in the
  *config*. It also takes the MkDocs `pages` configuration into account
  and converts the special `<< INFILE` syntax by copying them to the
  `gens_dir` as well.
  """

  for path in config['additional_search_paths']:
    path = os.path.abspath(path)
    sys.path.append(path)

  # Copy all template files from the source directory into our
  # generated files directory.
  log('Started copying source files...')
  for root, dirs, files in os.walk(config['docs_dir']):
    rel_root = os.path.relpath(root, config['docs_dir'])
    for fname in files:
      dest_fname = os.path.join(config['gens_dir'], rel_root, fname)
      makedirs(os.path.dirname(dest_fname))
      shutil.copyfile(os.path.join(root, fname), dest_fname)

  if 'pages' not in config:
    if pages_required:
      raise RuntimeError('pydocmd.yml does not have defined pages!')

    return

  for page in config['pages']:
    process_pages(page, config['gens_dir'])


def new_project():
  with open('pydocmd.yml', 'w') as fp:
    fp.write('site_name: Welcome to pydoc-markdown\ngenerate:\npages:\n- Home: index.md << ../README.md\n')


def log(*args, **kwargs):
  kwargs.setdefault('file', sys.stderr)
  print(*args, **kwargs)


def main():
  args = parser.parse_args()
  if args.command == 'new':
    new_project()
    return
  if args.command == 'simple' and not args.subargs:
    parser.error('need at least one argument')

  config = read_config() if args.command != 'simple' else default_config({})

  # Parse options.
  if args.command in ('generate', 'simple'):
    modspecs = []
    it = iter(args.subargs)
    while True:
      try: value = next(it)
      except StopIteration: break
      if value == '-c':
        try: value = next(it)
        except StopIteration: parser.error('missing value to option -c')
        key, value = value.partition('=')[::2]
        if value.startswith('['):
          if not value.endswith(']'):
            parser.error('invalid option value: {!r}'.format(value))
            value = value[1:-1].split(',')
        config[key] = value
      else:
        modspecs.append(value)
    args.subargs = modspecs

  loader = import_object(config['loader'])(config)
  preproc = import_object(config['preprocessor'])(config)

  if args.command != 'simple':
    mkdocs_exist = os.path.isfile('mkdocs.yml')
    copy_source_files(config, pages_required=not mkdocs_exist)

    if not mkdocs_exist:
      log('Generating temporary MkDocs config...')
      write_temp_mkdocs_config(config)

  # Build the index and document structure first, we load the actual
  # docstrings at a later point.
  log('Building index...')
  index = Index()

  def add_sections(doc, object_names, depth=1):
    if isinstance(object_names, list):
      [add_sections(doc, x, depth) for x in object_names]
    elif isinstance(object_names, dict):
      for key, subsections in object_names.items():
        add_sections(doc, key, depth)
        add_sections(doc, subsections, depth + 1)
    elif isinstance(object_names, str):
      # Check how many levels of recursion we should be going.
      expand_depth = len(object_names)
      object_names = object_names.rstrip('+')
      expand_depth -= len(object_names)

      def create_sections(name, level):
        if level > expand_depth:
          return
        index.new_section(doc, name, depth=depth + level, header_type=config.get('headers', 'html'))
        sort_order = config.get('sort')
        if sort_order not in ('line', 'name'):
          sort_order = 'line'
        need_docstrings = 'docstring' in config.get('filter', ['docstring'])
        for sub in dir_object(name, sort_order, need_docstrings):
          sub = name + '.' + sub
          sec = create_sections(sub, level + 1)

      create_sections(object_names, 0)
    else:
      raise RuntimeError(object_names)

  # Make sure that we can find modules from the current working directory,
  # and have them take precedence over installed modules.
  sys.path.insert(0, '.')

  if args.command == 'simple':
    # In simple mode, we generate a single document from the import
    # names specified on the command-line.
    doc = index.new_document('main.md')
    add_sections(doc, args.subargs)
  else:
    for pages in config.get('generate') or []:
      for fname, object_names in pages.items():
        doc = index.new_document(fname)
        add_sections(doc, object_names)

  preproc.link_lookup = {}
  for file, doc in index.documents.items():
    for section in doc.sections:
      preproc.link_lookup[section.identifier] = file
  # Load the docstrings and fill the sections.
  log('Started generating documentation...')
  for doc in index.documents.values():
    for section in filter(lambda s: s.identifier, doc.sections):
      loader.load_section(section)
      preproc.preprocess_section(section)

  if args.command == 'simple':
    for section in doc.sections:
      section.render(sys.stdout)
    return 0

  # Write out all the generated documents.
  for fname, doc in index.documents.items():
    fname = os.path.join(config['gens_dir'], fname)
    makedirs(os.path.dirname(fname))
    with open(fname, 'w') as fp:
      for section in doc.sections:
        section.render(fp)

  if args.command == 'generate':
    return 0

  log("Running 'mkdocs {}'".format(args.command))
  sys.stdout.flush()

  args = ['mkdocs', args.command] + args.subargs
  try:
    return subprocess.call(args)
  except KeyboardInterrupt:
    return signal.SIGINT


if __name__ == '__main__':
  main()




def function_without_docstrings(foo):
  pass


def function_with_docstring_on_same_line():
  """This is a pretty cool function.

  # Arguments
  None actually.

  # Example
  ```python
  # This is a very simple example.
  mycoolfunction(samplesize, width, **options)
  ```

  See also: #function_without_docstrings().
  See also: [this page](https://keras.io/layers/convolutional/#conv2d) for
    the list of possible arguments.
  """


def mycoolfunction(samplesize, width=32, **options):
  """
  This is a pretty cool function.

  # Arguments
  samplesize (int): Test sample size.
  width (int): The width.
  options: Additional options

  # Example
  ```python
  # This is a very simple example.
  mycoolfunction(samplesize, width, **options)
  ```

  See also: #myothercoolfunction().
  """

  pass


def myothercoolfunction(prettycool, url):
  """
  Don't you think?

      # This is part of a code block and will not be transformed.
      Code here

  ```
  # This is also part of a code block and will not be transformed.
  More code here
  ```

  # Parameters
  prettycool (any): Some parameter.
  url (string): the url for this thing (default: 'http://localhost#foobar')
  """

  pass

def add(a, b):
  """ Add two numbers.

  # Arguments

  a (int): First number
  b (int): Second number

  # Example

      assert add(2, 3) == 5

  Simple as that.

  Worth checking out: https://en.wikipedia.org/wiki/Addition#Properties
  """

  return a + b


class Breakfast(object):
  """
  This is a very simple class.

  # Arguments

  spam (Spam): 200g of spam
  eggs (Egg): 3 eggs
  ham (Ham): As much ham as you like.
  """

  def __init__(self, spam, eggs, ham=None):
    pass

  def __call__(self):
    """
    This is #Breakfast.__call__().
    """

  def cook(self):
    """
    Cooks the spam.

    # Raises

    OvercookError: If it cooked for too long and burned.
    """

  @property
  def price(self):
    " The price is hot. "
    return 33


b = Breakfast(True, False)


def rest_function(a, b, c):
  """
  This function is documented using ReST Syntax.

  :param a: The first parameter.
  :param b: The second parameter.
  :param c: The third parameter.
  :raise RuntimeError: Maybe sometimes.
  :return: Not much, really.
  """


class ClassWithoutDocs(object):

  def dosomething(self):
    " Abc. "

  @property
  def someprop(self):
    " Some property. "

  @classmethod
  def a_classmethod(self):
    " Test. "

  @staticmethod
  def a_staticmethod():
    pass



# Copyright (c) 2017  Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
This module provides utilities for importing Python objects by name.
"""

import types
import inspect


def import_module(name):
  """
  Imports a Python module assuming that the whole *name* identifies only a
  Python module and no symbol inside a Python module.
  """

  # fromlist must not be empty so we get the bottom-level module rather than
  # the top-level module.
  return __import__(name, fromlist=[''])


def import_object(name):
  """
  Like #import_object_with_scope() but returns only the object.
  """

  return import_object_with_scope(name)[0]


def import_object_with_scope(name):
  """
  Imports a Python object by an absolute identifier.

  # Arguments
  name (str): The name of the Python object to import.

  # Returns
  (any, Module): The object and the module that contains it. Note that
    for plain modules loaded with this function, both elements of the
    tuple may be the same object.
  """

  # Import modules until we can no longer import them. Prefer existing
  # attributes over importing modules at each step.
  parts = name.split('.')
  current_name = parts[0]
  obj = import_module(current_name)
  scope = None
  for part in parts[1:]:
    current_name += '.' + part
    try:
      if hasattr(obj, '__dict__'):
        # Using directly __dict__ for descriptors, where we want to get the descriptor's instance
        # and not calling the descriptor's __get__ method.
        sub_obj = obj.__dict__[part]
      else:
        sub_obj = getattr(obj, part)

      scope, obj = obj, sub_obj
    except (AttributeError, KeyError):
      try:
        obj = scope = import_module(current_name)
      except ImportError as exc:
        if 'named {}'.format(part) in str(exc):
          raise ImportError(current_name)
        raise
  return obj, scope


def force_lazy_import(name):
  """
  Import any modules off of "name" by iterating a new list rather than a generator so that this
  library works with lazy imports.
  """
  obj = import_object(name)
  module_items = list(getattr(obj, '__dict__', {}).items())
  for key, value in module_items:
    if getattr(value, '__module__', None):
      import_object(name + '.' + key)


def dir_object(name, sort_order, need_docstrings=True):
  prefix = None
  obj = import_object(name)
  if isinstance(obj, types.ModuleType):
    prefix = obj.__name__
  all = getattr(obj, '__all__', None)

  # Import any modules attached to this object so that this will work with lazy imports.  Otherwise
  # the block below will fail because the object will change while it's being iterated.
  force_lazy_import(name)

  by_name = []
  by_lineno = []
  for key, value in getattr(obj, '__dict__', {}).items():
    if isinstance(value, (staticmethod, classmethod)):
      value = value.__func__
    if key.startswith('_'):
      continue
    if not hasattr(value, '__doc__'):
      continue

    # If we have a type, we only want to skip it if it doesn't have
    # any documented members.
    if not (isinstance(value, type) and dir_object(name + '.' + key, sort_order, True)):
      if need_docstrings and not value.__doc__:
        continue
      if all is not None and key not in all:
        continue

    if prefix is not None and getattr(value, '__module__', None) != prefix:
      continue
    if sort_order == 'line':
      try:
        by_lineno.append((key, inspect.getsourcelines(value)[1]))
      except Exception:
        # some members don't have (retrievable) line numbers (e.g., properties)
        # so fall back to sorting those first, and by name
        by_name.append(key)
    else:
      by_name.append(key)
  by_name = sorted(by_name, key=lambda s: s.lower())
  by_lineno = [key for key, lineno in sorted(by_lineno, key=lambda r: r[1])]

  return by_name + by_lineno




import textwrap
from .utils import PreprocessorTest
from .test_processor_rst import RSTPreprocessorTest
from .test_processor_google import GooglePreprocessorTest
from pydocmd.preprocessors.smart import Preprocessor


class SmartRSTPreprocessorTest(RSTPreprocessorTest):
  preprocessor = Preprocessor()


class SmartGooglePreprocessorTest(GooglePreprocessorTest):
  preprocessor = Preprocessor()



from pydocmd.preprocessors.rst import Preprocessor as RSTPreprocessor
from pydocmd.preprocessors.google import Preprocessor as GooglePreprocessor


class Preprocessor(object):
  """
  This class implements the preprocessor for restructured text and google.
  """
  def __init__(self, config=None):
    self.config = config
    self._google_preprocessor = GooglePreprocessor(config)
    self._rst_preprocessor = RSTPreprocessor(config)

  def is_google_format(self, docstring):
    """
    Check if `docstring` is written in Google docstring format

    https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html
    """

    google_section_names = self._google_preprocessor.get_section_names()
    for section_name in google_section_names:
      if section_name in docstring:
        return True
    return False

  def preprocess_section(self, section):
    """
    Preprocessors a given section into it's components.
    """

    if self.is_google_format(section.content):
      return self._google_preprocessor.preprocess_section(section)

    return self._rst_preprocessor.preprocess_section(section)



# Copyright (c) 2017  Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
This module provides implementations to load documentation information from
an identifier as it is specified in the `pydocmd.yml:generate` configuration
key. A loader basically takes care of loading the documentation content for
that name, but is not supposed to apply preprocessing.
"""

from __future__ import print_function
from .imp import import_object_with_scope
import collections
import inspect
import types
import uuid
from yapf.yapflib.yapf_api import FormatCode

function_types = (types.FunctionType, types.LambdaType, types.MethodType,
  types.BuiltinFunctionType, types.BuiltinMethodType)
if hasattr(types, 'UnboundMethodType'):
  function_types += (types.UnboundMethodType,)


NotSet = object()


def trim(docstring):
  if not docstring:
    return ''
  lines = [x.rstrip() for x in docstring.split('\n')]
  lines[0] = lines[0].lstrip()

  indent = None
  for i, line in enumerate(lines):
    if i == 0 or not line: continue
    new_line = line.lstrip()
    delta = len(line) - len(new_line)
    if indent is None:
      indent = delta
    elif delta > indent:
      new_line = ' ' * (delta - indent) + new_line
    lines[i] = new_line

  return '\n'.join(lines)


class PythonLoader(object):
  """
  Expects absolute identifiers to import with #import_object_with_scope().
  """
  def __init__(self, config):
    self.config = config

  def load_section(self, section):
    """
    Loads the contents of a #Section. The `section.identifier` is the name
    of the object that we need to load.

    # Arguments
      section (Section): The section to load. Fill the `section.title` and
        `section.content` values. Optionally, `section.loader_context` can
        be filled with custom arbitrary data to reference at a later point.
    """

    assert section.identifier is not None
    obj, scope = import_object_with_scope(section.identifier)

    if '.' in section.identifier:
      default_title = section.identifier.rsplit('.', 1)[1]
    else:
      default_title = section.identifier

    section.title = getattr(obj, '__name__', default_title)
    section.content = trim(get_docstring(obj))
    section.loader_context = {'obj': obj, 'scope': scope}

    strip_self_param = inspect.isclass(scope)
    if isinstance(obj, (staticmethod, classmethod)):
      strip_self_param = strip_self_param and not isinstance(obj, staticmethod)
      obj = obj.__func__

    # Add the function signature in a code-block.
    if callable(obj):
      owner_class = scope if inspect.isclass(scope) else None
      sig = get_function_signature(obj, owner_class, strip_self_param)
      section.content = '```python\n{}\n```\n'.format(sig.strip()) + section.content


def get_docstring(function):
  if isinstance(function, (staticmethod, classmethod)):
    return function.__func__.__doc__ or ''
  elif hasattr(function, '__name__') or isinstance(function, property):
    return function.__doc__ or ''
  elif hasattr(function, '__call__'):
    return function.__call__.__doc__ or ''
  elif isinstance(function, type):
    return function.__doc__ or ''
  else:
    return ''


def get_full_arg_spec(func):
  if hasattr(inspect, 'getfullargspec'):
    spec = inspect.getfullargspec(func)
    return {k: getattr(spec, k) for k in dir(spec) if not k.startswith('_')}
  spec = inspect.getargspec(func)
  return {
    'args': spec.args,
    'varargs': spec.varargs,
    'varkw': spec.keywords,
    'defaults': spec.defaults,
    'kwonlyargs': [],
    'kwonlydefaults': [],
    'annotations': {}
  }


class Parameter(object):
  POS = 'POS'
  KWONLY = 'KWONLY'
  VARARGS_POS = 'VARARGS_POS'
  VARARGS_KW = 'VARARGS_KW'

  Value = collections.namedtuple('Value', 'value')

  def __init__(self, kind, name, annotation=None, default=None):
    self.kind = kind
    self.name = name
    self.annotation = annotation
    self.default = default

  def __repr__(self):
    return 'Parameter(kind={}, name={!r}, annotation={!r}, default={!r})'\
        .format(self.kind, self.name, self.annotation, self.default)

  def __str__(self):
    result = self.name
    if self.annotation:
      result += ': ' + repr(self.annotation.value)
    if self.default:
      result += ' = ' + repr(self.default.value)
    if self.kind == self.VARARGS_POS:
      result = '*' + result
    if self.kind == self.VARARGS_KW:
      result = '**' + result
    return result

  def replace(self, name=NotSet, annotation=NotSet,
              default=NotSet, kind=NotSet):
    return Parameter(
      self.kind if kind is NotSet else kind,
      self.name if name is NotSet else name,
      self.annotation if annotation is NotSet else annotation,
      self.default if default is NotSet else default)


def get_paramaters_from_arg_spec(argspec, strip_self=False):
  args = [Parameter(Parameter.POS, x) for x in argspec['args'][:]]
  if argspec['defaults']:
    offset = len(args) - len(argspec['defaults'])
    for i, default in enumerate(argspec['defaults']):
      args[i + offset] = args[i + offset].replace(default=Parameter.Value(default))

  if argspec['varargs']:
    args.append(Parameter(Parameter.VARARGS_POS, argspec['varargs']))

  for name in argspec['kwonlyargs']:
    args.append(Parameter(Parameter.KWONLY, name))

  if argspec['varkw']:
    args.append(Parameter(Parameter.VARARGS_KW, argspec['varkw']))

  if strip_self and args and args[0].name in ('cls', 'self'):
    args.pop(0)

  for i, param in enumerate(args):
    if param.name in argspec['annotations']:
      annotation = argspec['annotations'][param.name]
      args[i] = param.replace(annotation=Parameter.Value(annotation))

  return args


def format_parameters_list(parameters):
  found_kwonly = False
  result = []
  for param in parameters:
    if param.kind == Parameter.KWONLY and not found_kwonly:
      found_kwonly = True
      result.append('*')
    result.append(str(param))
  return ', '.join(result)


def get_function_signature(
  function,
  owner_class=None,
  strip_self_param=False,
  show_module=False,
  pretty=True,
):
  # Get base name.
  name_parts = []
  if show_module:
    name_parts.append(function.__module__)
  if owner_class:
    name_parts.append(owner_class.__name__)
  if hasattr(function, '__name__'):
    name_parts.append(function.__name__)
  else:
    name_parts.append(type(function).__name__)
    name_parts.append('__call__')
    function = function.__call__
  name = '.'.join(name_parts)

  try:
    argspec = get_full_arg_spec(function)
  except TypeError:
    parameters = []
  else:
    parameters = get_paramaters_from_arg_spec(argspec, strip_self=strip_self_param)

  # Prettify annotations.
  class repr_str(str):
    def __repr__(self):
      return self
  def prettify(val):  # type: (Parameter.Value) -> Parameter.Value
    if isinstance(val.value, type):
      val = Parameter.Value(repr_str(val.value.__name__))
    return val
  for i, param in enumerate(parameters):
    if param.annotation:
      param = param.replace(annotation=prettify(param.annotation))
    if param.default:
      param = param.replace(default=prettify(param.default))
    parameters[i] = param

  if pretty:
    # Replace annotations and defaults with placeholders that are valid syntax.
    supplements = {}
    counter = [0]

    def _add_supplement(value):
      annotation_id = '_{}'.format(counter[0])
      annotation_id += '_' * (len(repr(value)) - len(annotation_id))
      supplements[annotation_id] = value
      counter[0] += 1
      return repr_str(annotation_id)

    for i, param in enumerate(parameters):
      if param.annotation:
        param = param.replace(
          annotation=Parameter.Value(_add_supplement(param.annotation.value)))
      if param.default:
        param = param.replace(
          default=Parameter.Value(_add_supplement(param.default.value)))
      parameters[i] = param

  # Use a placeholder in pretty mode as the generated *name* may also
  # sometimes not be valid syntax (eg. if it includes the class name).
  name_placeholder = '_PH_' + '_' * (len(name)-2)
  sig = (name_placeholder if pretty else name) + '(' + format_parameters_list(parameters) + ')'

  if pretty:
    sig, _ = FormatCode('def ' + sig + ': pass', style_config='pep8')
    sig = sig[4:].rpartition(':')[0]

    # Replace the annotation and default placeholders with the actual values.
    for placeholder, annotation in supplements.items():
      sig = sig.replace(placeholder, repr(annotation))

    # Replace the placeholder and fix indents.
    sig = sig.replace(name_placeholder, name)
    delta = len(name_placeholder) - len(name)
    lines = sig.split('\n')
    for i, line in enumerate(lines[1:], 1):
      indent = len(line) - len(line.lstrip()) - delta - 4  # 4 for "def "
      if indent <= 0 and line.strip() != ')':
        indent = 4
      lines[i] = ' ' * indent + line.lstrip()
    sig = '\n'.join(lines)

  return sig




import io
import re
import setuptools
import sys

with io.open('pydocmd/__init__.py', encoding='utf8') as fp:
  version = re.search(r"__version__\s*=\s*'(.*)'", fp.read()).group(1)

with io.open('README.md', encoding='utf8') as fp:
  long_description = fp.read()

requirements = ['MkDocs>=1.0.0', 'Markdown>=2.6.11', 'PyYAML>=3.12', 'six>=0.11.0', 'yapf>=0.26.0']

setuptools.setup(
  name = 'pydoc-markdown',
  version = version,
  author = 'Niklas Rosenstein',
  author_email = 'rosensteinniklas@gmail.com',
  description = 'Create Python API documentation in Markdown format',
  long_description = long_description,
  long_description_content_type = 'text/markdown',
  url = 'https://github.com/NiklasRosenstein/pydoc-markdown',
  license = 'MIT',
  packages = setuptools.find_packages('.', ['test', 'test.*', 'docs', 'docs.*']),
  package_dir = {'': '.'},
  include_package_data = False,
  install_requires = requirements,
  extras_require = {},
  tests_require = [],
  python_requires = None, # TODO: '>=3.4,<4.0.0',
  data_files = [],
  entry_points = {
    'console_scripts': [
      'pydocmd=pydocmd.__main__:main',
    ]
  },
  cmdclass = {},
  keywords = ['markdown', 'pydoc', 'generator', 'docs', 'documentation'],
  classifiers = ['Development Status :: 3 - Alpha', 'Intended Audience :: Developers', 'Intended Audience :: End Users/Desktop', 'Topic :: Software Development :: Code Generators', 'Topic :: Utilities', 'License :: OSI Approved :: MIT License', 'Programming Language :: Python :: 2', 'Programming Language :: Python :: 2.7', 'Programming Language :: Python :: 3', 'Programming Language :: Python :: 3.3', 'Programming Language :: Python :: 3.4', 'Programming Language :: Python :: 3.5'],
)
